import argparse
import sys
from os import path, mkdir, chdir
from shutil import copyfile

from git import Repo


def run():
    parser = argparse.ArgumentParser(prog='pastry init')
    parser.add_argument('directory', type=str, help='Path to directory to create (or default to current directory)')
    args = parser.parse_args(sys.argv[2:])  # type: argparse.Namespace

    if args.directory is not None:
        mkdir(args.directory)
        chdir(args.directory)

    repo = Repo.init('.')  # type:  git.repo.base.Repo
    for filename in ['pie.py', '.gitignore']:
        source_file_path = path.join(path.dirname(__file__), '..', '..', 'data', filename)
        copyfile(source_file_path, filename)
        repo.index.add([filename])




from datetime import datetime
from types import ModuleType

import numpy as np
import pandas as pd
from git import Repo
from joblib import Parallel, delayed
from pypastry.experiment import Experiment
from pypastry.experiment.hasher import get_dataset_hash
from pypastry.experiment.results import ResultsRepo
from scipy.stats import sem
from sklearn.base import BaseEstimator, is_classifier, clone
from sklearn.model_selection import check_cv


class ExperimentRunner:
    def __init__(self, git_repo: Repo, results_repo: ResultsRepo, results_display: ModuleType):
        self.git_repo = git_repo
        self.results_repo = results_repo
        self.results_display = results_display

    def run_experiment(
        self,
        experiment: Experiment,
        force: bool,
        message: str,
        limit: int
    ):
        print("Got dataset with {} rows".format(len(experiment.dataset)))
        if force or self.git_repo.is_dirty():
            print("Running evaluation")
            self._run_evaluation(experiment, message)
            results = self.results_repo.get_results(self.git_repo)
            self.results_display.cache_display(results)
        else:
            print("Clean repo, nothing to do")
        self.results_display.print_cache_file(limit)

    def _run_evaluation(self, experiment: Experiment, message: str):
        X = experiment.dataset.drop(columns=[experiment.label_column])
        y = experiment.dataset[experiment.label_column]
        predictors = [experiment.predictor]
        if experiment.group_column is None:
            groups = None
        else:
            groups = X[experiment.group_column]
            X = X.drop(columns=[experiment.group_column])
        run_infos = _evaluate_predictors(X, predictors, y, groups, experiment.cross_validator, experiment.scorer)
        self.git_repo.git.add(update=True)
        dataset_hash = get_dataset_hash(experiment.dataset)
        dataset_info = {
            'hash': dataset_hash,
            'columns': experiment.dataset.columns.tolist(),
        }
        new_filenames = self.results_repo.save_results(run_infos, dataset_info)
        self.git_repo.index.add(new_filenames)
        self.git_repo.index.commit(message)


def _evaluate_predictors(X, predictors, y, groups, cross_validator, scorer):
    run_infos = []
    for predictor in predictors:
        start = datetime.utcnow()

        scores = _get_scores(X, y, groups, cross_validator, predictor, scorer)

        end = datetime.utcnow()

        scores_array = np.hstack(scores)
        mean_score = scores_array.mean()
        sem_score = sem(scores_array)
        results = {'test_score': mean_score, 'test_score_sem': sem_score}

        model_info = get_model_info(predictor)

        run_info = {
            'run_start': str(start),
            'run_end': str(end),
            'run_seconds': (end - start).total_seconds(),
            'results': results,
            'model_info': model_info,
        }
        run_infos.append(run_info)
    return run_infos


def get_model_info(model: BaseEstimator):
    info = model.get_params()
    info['type'] = type(model).__name__
    return info


def _get_scores(X, y, groups, cv, estimator, scorer):
    cv = check_cv(cv, y, classifier=is_classifier(estimator))

    # We clone the estimator to make sure that all the folds are
    # independent, and that it is pickle-able.
    parallel = Parallel(n_jobs=None, verbose=False,
                        pre_dispatch='2*n_jobs')
    scores = parallel(
        delayed(_fit_and_predict)(
            clone(estimator), X, y, train, test, groups, scorer)
        for train, test in cv.split(X, y, groups))
    return scores


def _fit_and_predict(estimator: BaseEstimator, X, y, train, test, groups, scorer):
    if groups is not None:
        return _fit_and_predict_groups(X, estimator, groups, scorer, test, train, y)
    else:
        return _fit_and_predict_simple(X, estimator, scorer, test, train, y)


def _fit_and_predict_simple(X, estimator, scorer, test, train, y):
    X_train = X.iloc[train]
    y_train = y.iloc[train]
    estimator.fit(X_train, y_train)
    X_test = X.iloc[test]
    y_test = y.iloc[test]
    score = scorer(estimator, X_test, y_test)
    return [score]


def _fit_and_predict_groups(X, estimator, groups, scorer, test, train, y):
    X_train = X.iloc[train]
    y_train = y.iloc[train]
    estimator.fit(X_train, y_train)
    X_test = X.iloc[test]
    y_test = y.iloc[test]
    groups_test = groups.iloc[test]
    test_df = pd.DataFrame(X_test)
    test_df['y'] = y_test
    test_df['groups'] = groups_test
    # test_df = pd.DataFrame({'X': X_test.values, 'y': y_test.values, 'groups': groups_test.values})
    test_groups = test_df.groupby('groups')
    scores = []
    for key, group in test_groups:
        X_group = group[X.columns]
        score = scorer(estimator, X_group, group['y'])
        scores.append(score)
    print("SCORES", scores)
    return scores




import argparse
import sys

from git import Repo
from pypastry import display

from pypastry.experiment.evaluation import ExperimentRunner
from pypastry.experiment.results import ResultsRepo
from pypastry.paths import REPO_PATH, RESULTS_PATH


def run():
    parser = argparse.ArgumentParser(prog='pastry run')
    parser.add_argument('-f', '--force', action='store_true', help='Force a re-run of the experiment')
    parser.add_argument('-m', '--message', type=str, required=True, help='Git commit message')
    parser.add_argument('-l', '--limit', type=int, help='Limit lines to print')

    args = parser.parse_args(sys.argv[2:])

    sys.path.append('.')
    import pie
    experiment = pie.get_experiment()

    git_repo = Repo(REPO_PATH)  # type: pypastry.experiment.Experiment
    results_repo = ResultsRepo(RESULTS_PATH)  # type: pypastry.experiment.results.ResultsRepo
    runner = ExperimentRunner(git_repo, results_repo, display)  # type:
    # pypastry.experiment.evaluation.ExperimentRunner
    runner.run_experiment(experiment, args.force, args.message, args.limit)



from unittest.mock import Mock

from pandas import DataFrame
from sklearn.dummy import DummyClassifier
from sklearn.metrics import accuracy_score, make_scorer
from sklearn.model_selection import StratifiedShuffleSplit, GroupShuffleSplit
from sklearn.tree import DecisionTreeClassifier

from pypastry.experiment import Experiment
from pypastry.experiment.evaluation import ExperimentRunner


def test_simple_evaluation():
    # Given
    dataset = DataFrame({
        'a': [1, 1, 0, 0],
        'b': [1, 1, 0, 0],
    })

    cross_validation = StratifiedShuffleSplit(n_splits=1, test_size=0.5)
    predictor = DecisionTreeClassifier()
    scorer = make_scorer(accuracy_score)
    experiment = Experiment(dataset, 'b', predictor, cross_validation, scorer)

    git_mock = Mock()
    results_repo_mock = Mock()
    new_results_files = ['results/abc.json']
    results_repo_mock.save_results.return_value = new_results_files
    results_display_mock = Mock()
    runner = ExperimentRunner(git_mock, results_repo_mock, results_display_mock)
    commit_message = "Test commit message"

    # When
    runner.run_experiment(experiment, False, commit_message)

    # Then
    call_args_list = results_repo_mock.save_results.call_args_list
    assert 1 == len(call_args_list)
    run_infos, dataset_info = call_args_list[0][0]

    assert len(run_infos) == 1

    results = run_infos[0]['results']
    assert 1.0 == results['test_score']
    assert ['a', 'b'] == dataset_info['columns']

    # TODO: check the hash. Need to find a way to make this consistent between python versions etc.
    # assert '28ea628a50a47c726a9b0ec437c88fc4742d81fd' == dataset_info['hash']

    git_mock.git.add.assert_called_once_with(update=True)
    git_mock.index.add.assert_called_once_with(new_results_files)
    git_mock.index.commit.assert_called_once_with(commit_message)

    assert 1 == len(results_display_mock.cache_display.call_args_list)
    print(results_display_mock.cache_display.call_args[0])
    assert len(results_display_mock.cache_display.call_args[0]) > 0
    assert 1 == len(results_display_mock.print_cache_file.call_args_list)


def test_grouped_evaluation():
    label = [i % 2 for i in range(100)]
    dataset = DataFrame({
        'a': label,
        'b': label,
        'g': [i // 2 for i in range(100)]
    })

    cross_validation = GroupShuffleSplit(n_splits=1, test_size=0.5)
    predictor = DummyClassifier(strategy='constant', constant=1)
    scorer = make_scorer(accuracy_score)
    experiment = Experiment(dataset, 'b', predictor, cross_validation, scorer, group_column='g')

    git_mock = Mock()
    results_repo_mock = Mock()
    results_display_mock = Mock()
    runner = ExperimentRunner(git_mock, results_repo_mock, results_display_mock)

    runner.run_experiment(experiment, False, "Test commit message")

    assert 1 == len(results_repo_mock.save_results.call_args_list)
    run_infos, dataset_info = results_repo_mock.save_results.call_args[0]

    assert len(run_infos) == 1

    results = run_infos[0]['results']
    assert 0.5 == results['test_score']
    assert 0.0 == results['test_score_sem']



"""
Handle displaying results.

This code needs to display results really fast. That's why it has some odd things:
 - Inline imports
 - Lack of typing for the results_repo to avoid unnecessary imports

This is because I don't like having to wait a second for things to be imported.
I want my pastry now!
"""
import os
from typing import Any, Dict, List, Iterator, TYPE_CHECKING

from pypastry.paths import DISPLAY_PATH, DISPLAY_DIR
if TYPE_CHECKING:
    import pypastry


def cache_display(results_from_repo: Iterator['pypastry.experiment.results.Result']) -> None:
    from pandas import DataFrame

    results = []
    for result in results_from_repo:
        data = result.data
        result = {
            'Git hash': result.git_hash,
            'Dataset hash': data['dataset']['hash'][:8],
            'Run start': data['run_start'][:19],
            'Model': data['model_info']['type'],
            'Score': "{:.3f} ± {:.3f}".format(data['results']['test_score'],
                                              data['results']['test_score_sem']),
            'Duration (s)': "{:.2f}".format(data['run_seconds']),
        }
        results.append(result)
    results.sort(key=lambda row: row['Run start'])
    recent_results = results
    results_dataframe = DataFrame(recent_results)
    display = repr(results_dataframe)

    try:
        os.mkdir(DISPLAY_DIR)
    except FileExistsError:
        pass

    with open(DISPLAY_PATH, 'w') as output_file:
        output_file.write(display)


def print_cache_file(limit = False):
    with open(DISPLAY_PATH) as display_file:
        read_lines = display_file.read()
        read_list = read_lines.split("\n")
        if limit:
            limit = min(limit, len(read_list)-3)
        else:
            limit = len(read_list)-3
        print(read_list[0])
        print("\n".join(read_list[-(2+limit):-2]))
        print(f'Cache provides:\n{read_list[-1]}')



import hashlib
from io import BytesIO

import pandas as pd
from pandas import DataFrame

BLOCKSIZE = 65536


def get_dataset_hash(dataset: DataFrame):
    buffer = BytesIO()
    dataset.to_parquet(buffer)
    data = buffer.getvalue()

    return get_bytes_hash(data)


def get_bytes_hash(data: bytes):
    hasher = hashlib.sha1()
    for i in range(0, len(data), BLOCKSIZE):
        block = data[i:i + BLOCKSIZE]
        hasher.update(block)
    return hasher.hexdigest()


if __name__ == "__main__":
    data = pd.read_csv('../data/iris.csv')
    hash = get_dataset_hash(data)
    print(hash)



import json
import os
from glob import glob
from os import mkdir
from tempfile import NamedTemporaryFile
from typing import Dict, Any, List, NamedTuple

from git import Repo

Result = NamedTuple('Result', [('data', Dict[str, Any]), ('git_hash', str)])


class ResultsRepo:
    def __init__(self, results_path: str):
        self.results_path = results_path

    def save_results(self, run_infos: List[Dict[str, Any]], dataset_info: Dict[str, Any]) -> List[str]:
        try:
            mkdir(self.results_path)
        except FileExistsError:
            pass
        new_filenames = []
        for i, run_info in enumerate(run_infos):
            run_info['dataset'] = dataset_info
            with NamedTemporaryFile(mode='w', prefix='result-', suffix='.json',
                                    dir=self.results_path, delete=False) as output_file:
                json.dump(run_info, output_file, indent=4)
                output_file.flush()
            new_filenames.append(output_file.name)
        return new_filenames

    def get_results(self, git_repo):
        results_glob = os.path.join(self.results_path, '*')
        for path in glob(results_glob):
            with open(path) as results_file:
                git_hash = next(git_repo.iter_commits(paths=path)).hexsha[:8]
                result_json = json.load(results_file)
                yield Result(result_json, git_hash)



from pypastry.display import print_cache_file, cache_display
from pypastry.paths import RESULTS_PATH


def run():
    try:
        print_cache_file()
    except FileNotFoundError:
        from pypastry.experiment.results import ResultsRepo
        from git import Repo
        results_repo = ResultsRepo(RESULTS_PATH)
        results = results_repo.get_results(Repo('.'))
        cache_display(results)
        print_cache_file()



import sys

from setuptools import setup

if sys.version_info < (3, 5):
    sys.exit('Sorry, PyPastry requires Python version 3.5 or greater')

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pypastry',
    version='0.0.1',
    description='PyPastry machine learning experimentation framework',
    author='Daoud Clarke',
    url='https://github.com/datapastry/pypastry',
    scripts=['pastry'],
    install_requires=['tomlkit', 'pandas', 'scikit-learn', 'pyarrow', 'gitpython'],
    package_dir ={'': 'pypastry'},
    long_description = long_description,
    long_description_content_type="text/markdown",
    classifiers = [

        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',

)



from typing import Any

from pandas import DataFrame
from sklearn.base import BaseEstimator


class Experiment:
    def __init__(self, dataset: DataFrame, label_column: str, predictor: BaseEstimator,
                 cross_validator: Any, scorer, group_column=None):
        self.dataset = dataset
        self.label_column = label_column
        self.predictor = predictor
        self.cross_validator = cross_validator
        self.scorer = scorer
        self.group_column = group_column
